/******************************************************************************

	Camera Image Capturing and Recording System

	main.c

		2012.7.6   S.Suwa   http://www.suwa-koubou.jp


	MCLR        (1) 
	RA0(RP5)    (2) U2TX  for debug 
	RA1(RP6)    (3) U2RX  for debug
	RB0         (4) PGED1 for Pickit2
	RB1         (5) PGEC1 for Pickit2
	RB2         (6) Device Mode LED
    RB3         (7) Host Mode LED
	VSS         (8) GND
	RA2         (9) Mode Change Switch, pullup, via 0.1uF to GND
	RA3        (10) Mode Change Relay Control
	RB4        (11) (NC)
	RA4        (12) (NC)
	VDD        (13) 3V3
	RB5        (14) USB_BUS_SENSE to USB connector VBus Pin

	AVDD       (28) 3V3
	AVSS       (27) GND
	RB15       (26) SD Card Write Enable detect
	RB14       (25) SD Card Slot in detect
	RB13(RP13) (24) SPI MISO <-- SD Card data out
	VUSB       (23)
	D-         (22) USB connector D- pin
	D+         (21) USB connector D+ pin
	VCap       (20) via 10uF to GND
	DISVREG    (19) GND
	RB9(RP9)   (18) SPI CLK  --> SD Card clock
    RB8(RP8)   (17) SPI MOSI --> SD Card data in
	RB7        (16) SPI Slave Select --> SD Card CS
	VBUS       (15) 5V

******************************************************************************/

//#include <stdlib.h>
#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "usb_config.h"
//#include "USB/usb.h"
//#include "USB/usb_host_generic.h"
#include "usb.h"
#include "usb_host_generic.h"
#include "rtc.h"
//#include "timer.h"

#include "SD-SPI.h"
#include "FSIO.h"

#include "usb_function_msd.h"

#define	DEBUG_MODE
#ifndef	DEBUG_MODE
	#define	UART2PrintString(x)
	#define	UART2PutHex(x)
	#define	UART2PutChar(x)
	#define	UART2PutDec(x)
#endif

// *****************************************************************************
// *****************************************************************************
// Configuration Bits
// *****************************************************************************
// *****************************************************************************

_CONFIG1(WDTPS_PS1 & FWPSA_PR32 & WINDIS_OFF & FWDTEN_OFF & ICS_PGx1 & GWRP_OFF & GCP_OFF & JTAGEN_OFF)
_CONFIG2(POSCMOD_NONE & I2C1SEL_PRI & IOL1WAY_OFF & OSCIOFNC_ON & FCKSM_CSDCMD & FNOSC_FRCPLL & PLL96MHZ_ON & PLLDIV_DIV2 & IESO_OFF)
_CONFIG3(WPFP_WPFP0 & SOSCSEL_IO & WUTSEL_LEG & WPDIS_WPDIS & WPCFG_WPCFGDIS & WPEND_WPENDMEM)
//_CONFIG3(WPFP_WPFP0 & SOSCSEL_SOSC & WUTSEL_LEG & WPDIS_WPDIS & WPCFG_WPCFGDIS & WPEND_WPENDMEM)
_CONFIG4(DSWDTPS_DSWDTPS3 & DSWDTOSC_LPRC & RTCOSC_SOSC & DSBOREN_OFF & DSWDTEN_OFF)

// *****************************************************************************
// *****************************************************************************
// Constants
// *****************************************************************************
// *****************************************************************************
// FinePix Control Order
#define	FPIX_RESET			0	// fpix_command(FPIX_RESET);
#define	FPIX_FRAME_REQUEST	1	// fpix_command(FPIX_FRAME_REQUEST);

// *****************************************************************************
// *****************************************************************************
// Data Structures
// *****************************************************************************
// *****************************************************************************
// FinePix Camera States
typedef enum
{
	CAMERA_DETACHED = 0,
	CAMERA_ATTACHED,
} CAMERA_STATE;

// Application Mode
typedef enum
{
    HOST_MODE,
    DEVICE_MODE
} ACTIVE_MODE;


// *****************************************************************************
// *****************************************************************************
// Macros
// *****************************************************************************
// *****************************************************************************
#define	MODE_SWITCH		PORTAbits.RA2
#define	MODE_SELECT_RLY	LATAbits.LATA3

#define	DEVICE_MODE_LED	LATBbits.LATB2
#define	HOST_MODE_LED	LATBbits.LATB3
#define	_DI()			asm("DISI #200")
#define	_EI()			asm("DISI #0")

// *****************************************************************************
// *****************************************************************************
// Global Variables
// *****************************************************************************
// *****************************************************************************
static	ACTIVE_MODE 	active_mode;	// Host or Device 

// Device Mode variable
LUN_FUNCTIONS LUN[MAX_LUN + 1] = 
{
    {
        &MDD_SDSPI_MediaInitialize,
        &MDD_SDSPI_ReadCapacity,
        &MDD_SDSPI_ReadSectorSize,
        &MDD_SDSPI_MediaDetect,
        &MDD_SDSPI_SectorRead,
        &MDD_SDSPI_WriteProtectState,
        &MDD_SDSPI_SectorWrite
    }
};

/* Standard Response to INQUIRY command stored in ROM 	*/
const ROM InquiryResponse inq_resp = {
	0x00,	// peripheral device is connected, direct access block device
	0x80,	// removable
	0x04,	// version = 00=> does not conform to any standard, 4=> SPC-2
	0x02,	// response is in format specified by SPC-2
	0x20,	// n-4 = 36-4=32= 0x20
	0x00,	// sccs etc.
	0x00,	// bque=1 and cmdque=0,indicates simple queueing 00 is obsolete,
			// but as in case of other device, we are just using 00
	0x00,	// 00 obsolete, 0x80 for basic task queueing
	{'M','i','c','r','o','c','h','p'
    },
	// this is the T10 assigned Vendor ID
	{'M','a','s','s',' ','S','t','o','r','a','g','e',' ',' ',' ',' '
    },
	{'0','0','0','1'
    }
};

// Host mode variable
BYTE	deviceAddress;  	// Address of the device on the USB
BYTE	DataPacket[1024];	// Captured Image data receive buffer
CAMERA_STATE camera_state;	// camera state
WORD	capture_interval;	// unit second
BOOL	set_rtc_flg;		// real time clock, first setting flag

//******************************************************************************
//******************************************************************************
// Internal Function Prototype
//******************************************************************************
//******************************************************************************
void	do_host_mode(void);
void	do_device_mode(void);
void	InitializeSystem(void);

WORD	fpix_frame_read(BYTE deviceAddress, BYTE *buffer, DWORD length);
BOOL	fpix_command(BYTE deviceAddress, BYTE order);
int		one_frame_recording(FSFILE *fout);


//******************************************************************************
//******************************************************************************
// Main
//******************************************************************************
//******************************************************************************
int main ( void )
{
    // Initialize the processor and peripherals.
    InitializeSystem();

	// opening message
    UART2PrintString( "\r\n*** FinePix PC Camera Image Capture and Image Recording.\r\n" );

	// initialize status
	camera_state = CAMERA_DETACHED;
	active_mode = DEVICE_MODE;

	// main loop
	for(;;){
		if(active_mode == HOST_MODE){
			do_host_mode();
		}else if(active_mode == DEVICE_MODE){
			do_device_mode();
		}
	}

	return 0;
}

/*-----------------------------------------------------------------------------
	read mode change request switch
-----------------------------------------------------------------------------*/

ACTIVE_MODE	read_mode_switch(void)
{
	unsigned int	tm;
	// switch on ?
	if(MODE_SWITCH == 1) return active_mode;


	// wait 10msec
	tm = timer_val;	// count down at blink_led()
	while(tm == timer_val);
	if(MODE_SWITCH == 1) return active_mode;

	// changed mode
	if(active_mode == HOST_MODE)
		active_mode = DEVICE_MODE;
	else
		active_mode = HOST_MODE;

	// wait for switch off
	while(MODE_SWITCH == 0){
		;
	}

	return active_mode;
}


/*-----------------------------------------------------------------------------
	mode led turn on/off
-----------------------------------------------------------------------------*/

void	blink_led(void)
{
	if(timer_val == 0){
		if(active_mode == HOST_MODE){
			HOST_MODE_LED = ~HOST_MODE_LED;
			DEVICE_MODE_LED = 0;
		}else{
			DEVICE_MODE_LED = ~DEVICE_MODE_LED;
			HOST_MODE_LED = 0;
		}
		_DI();
		timer_val = 20;	// LED 200msec interval turn on, turn off
		_EI();
	}
}


/*=============================================================================
	Device Mode, SD-Card reader, Mass Strage Device
=============================================================================*/

void	do_device_mode(void)
{
	UART2PrintString("do_device_mode()\r\n");
	// USB signal line change to Device.(Relay off)
	MODE_SELECT_RLY = 0;

	// check BUS SENSE
	if(USB_BUS_SENSE == 0){
		active_mode = HOST_MODE;
		UART2PrintString("Not Ready USB BUS. Return to HOST mode.\r\n");
		return;
	}

	// set OTG module to USB Device(MSD)
	USBDeviceInit();
	USBDeviceTasks();
	UART2PrintString("complete device init.\r\n");

	// initilize SPI
	MDD_SDSPI_InitIO();
	USBDeviceTasks();
	UART2PrintString("complete SDSPI init.\r\n");

	while(USB_BUS_SENSE){

		// wait for first device attach
		while(USBDeviceState == DETACHED_STATE){
			// LED blinking during detached state
			USBDeviceTasks();
			blink_led();
			// request mode change 
			if(read_mode_switch() != DEVICE_MODE) return;
		}

		// *** ATTACHED to PC
		// LED turn on
		DEVICE_MODE_LED = 1;	// on
		HOST_MODE_LED = 0;		// off

		// do SD-Card reader
		while(USBDeviceState != DETACHED_STATE){
			USBDeviceTasks();
   	 		if((USBDeviceState < CONFIGURED_STATE)||(USBSuspendControl==1))
				continue;
	   		MSDTasks();    
		}
	}

	active_mode = HOST_MODE;
	return;
}


/*=============================================================================
	Host Mode, FinePix PC Camera Image capturing and recording
=============================================================================*/
static int	file_number = 1;

void	do_host_mode(void)
{
	BOOL	FSReady;
	FSFILE	*fout, *fin;
	int		len, ret;
	char	buf[32];

	UART2PrintString("do_host_mode()\r\n");
	// USB signal line change to Host.(Relay on)
	MODE_SELECT_RLY = 1;

	// set OTG module to USB Host(GENERIC)
	USBHostInit(0);
	UART2PrintString("Host init.\r\n");

	// wait for first camera attach
	while(camera_state == CAMERA_DETACHED){
		// check B receptacle Vbus.
		if(USB_BUS_SENSE){
			active_mode = DEVICE_MODE;
			return;
		}
		// LED blinking during detached state
		USBHostTasks();
		blink_led();
		// request mode change 
		if(read_mode_switch() != HOST_MODE) return;
	}

	// LED turn on
	HOST_MODE_LED = 1;
	DEVICE_MODE_LED = 0;

	// camera attached!! do camera reset
	fpix_command(deviceAddress, FPIX_RESET);

	// wait 1sec
	_DI();
	timer_val = 100;
	_EI();
	while(timer_val);


	// initialize Fat FileSystem 
	FSReady = NULL;
	while(!FSReady){
		FSReady = FSInit();
		USBHostTasks();
	}
	UART2PrintString("FS complete.\r\n");

	// open camera.ini file in root directory
	if(!set_rtc_flg && (fin = FSfopen("CAMERA.INI", "r")) != NULL){
		/*
		 * camera.ini
		 *   first line: YYYY/MM/DD hh:mm:ss xxxxxx\r\n
	     *      xxxx is frame read interval time, unit 1 second.
		 */
		len = FSfread(buf, 1, 32, fin);
		if(len > 22){
			buf[19] = ' ';
			rtc_set(buf);
			capture_interval = atoi(buf+20);
			set_fattime();
			set_rtc_flg = 1;
		}
		FSfclose(fin);
	}

	// changed "data" directory
	if(FSchdir("DATA") == 0) UART2PrintString("set current is DATA directory.\r\n");

	// unset capture_interval?
	if(!set_rtc_flg) capture_interval = 5;

	// limitation capture interval time
	if(capture_interval < 1) capture_interval = 1;	// minimum 3sec
	if(capture_interval > 43200) capture_interval = 43200;	// maximum 12Hour

	UART2PrintString("capture_interval=");
	UART2PutDec(capture_interval);
	UART2PrintString("\r\n");


	// set capture timing 	
	set_capture_time(capture_interval);

	while(camera_state == CAMERA_ATTACHED){

		// file open
		set_fattime();
		sprintf(buf, "IMG%05d.JPG", file_number++);
		fout = FSfopen(buf, "w");
		if(fout == NULL){
			UART2PrintString(buf); UART2PrintString("Open error. code=");
			UART2PutHex(FSerror()); UART2PrintString("\r\n");
			while(1){
				blink_led();
				// request mode change 
				if(read_mode_switch() != HOST_MODE) return;
			}
		}

		// frame capture
		ret = one_frame_recording(fout);

		// file close
		set_fattime();
		FSfclose(fout);

		// if capture error, then file delete.
		if(ret == -1){
			FSremove(buf);
		}

		#ifdef DEBUG_MODE
		if(ret == 0){
			UART2PrintString(buf); UART2PrintString(" write!\r\n");
		}else{
			UART2PrintString(buf); UART2PrintString(" capture error. deleted file!\r\n");
		}
		#endif
		// wait time until next frame capture
		while(!is_capture && camera_state == CAMERA_ATTACHED){
			USBHostTasks();
		}
		// clear capture timing flag
		_DI();
		is_capture = 0;
		_EI();

	}//while


	// Now Clock Time Store in to CAMERA.INI
	get_rtcstr(buf);
	sprintf(buf+19, " %d\r\n", capture_interval);
	if(FSchdir("\\") == 0){
		UART2PrintString("chdir to root directory.\r\n");
		if((fout = FSfopen("CAMERA.INI", "w")) != NULL){
			FSfwrite(buf, 1, 32, fout);
			FSfclose(fout);
			UART2PrintString("Now Clock Time Store in to CAMERA.INI\r\n");
			UART2PrintString(" store time="); UART2PrintString(buf);
		}else{
			UART2PrintString("CAMERA.INI write open error!\r\n");
		}
	}else{
		UART2PrintString("chdir error!\r\n");
	}	

	// change to Device Mode
	active_mode = DEVICE_MODE;
	return;
}


/*------------------------------------------------------------------------
	one frame capture and recording to file

	return code  0 success
                -1 error, fail
------------------------------------------------------------------------*/

int		one_frame_recording(FSFILE *fout)
{
	WORD	size;

	// frame capture start command
	if(fpix_command(deviceAddress, FPIX_FRAME_REQUEST) == FALSE){
		UART2PrintString("fpix_command error!\r\n");
		return -1;	// error !!
	}

	while(camera_state == CAMERA_ATTACHED){
		if((size = fpix_frame_read(deviceAddress, DataPacket, sizeof(DataPacket))) == 0){
			UART2PrintString("fpix_frame_read error!\r\n");
			return -1;	// error !!
		}

		FSfwrite(DataPacket, size, 1, fout);

		// check last packet
		if( (size < sizeof(DataPacket)) || ( DataPacket[size -2] == 0xff && DataPacket[size-1] == 0xd9)) {
			return 0;						
		}
	}

	return -1;

}


/*========================================================================
	InitializeSystem()
========================================================================*/

void	InitializeSystem (void)
{
	unsigned int pll_startup_counter;

	// Fosc(system clock) 8MHz postscale = 1:1 x4 = 32MHz, Fcy = Fosc/2
	CLKDIV = 0;

	// PLL clock start up
	pll_startup_counter = 600;
	CLKDIVbits.PLLEN = 1;
	while(pll_startup_counter--);

	//--- setting interrupt ---------------------------------------
	INTCON1 = 0;			// NSTDIS=0 nesting ok

	//--- setting I/O port ----------------------------------------
	AD1PCFG = 0xffff;	// Analog disable
	LATA = 0x0000; TRISA = 0xffff;	// RA are default input.
	LATB = 0x0000; TRISB = 0xffff;	// RB are default input.

	TRISAbits.TRISA0 = 0;	// RA0(UxTX) is output
	TRISAbits.TRISA1 = 1;	// RA1(UxRX) is input
	LATAbits.LATA0 = 1;		// RA0(UxTX) to level high

	TRISAbits.TRISA2 = 1;	// Host/Device Mode, Change Switch Input Pin
	TRISAbits.TRISA3 = 0;	// USB D+,D- line Host/Device Change Relay
	LATAbits.LATA3 = 0;		//    Select Device

	TRISBbits.TRISB2 = 0;	// Device Mode LED drive Output Pin
	TRISBbits.TRISB3 = 0;	// Host Mode LED drive Output Pin
	LATBbits.LATB2 = 0;		// turn off Device Mode LED
	LATBbits.LATB3 = 0;		// turn off Host Mode LED

	TRISBbits.TRISB5 = 1;	// BUS sense Input Pin, in Device Mode 

	//TRISBbits.TRISB7 = 1;	// SD Card Slot in detect
	//TRISBbits.TRISB8 = 1;	// SD Card Write Enable detect
	//TRISBbits.TRISB9 = 0;	// SPI Slave Select --> SD Card CS
	//TRISBbits.TRISB13 = 0;	// SPI clock        --> SD Card clk
	//TRISBbits.TRISB14 = 0;	// SPI data output  --> SD Card in
	//TRISBbits.TRISB15 = 1;	// SPI data input   <-- SD Card out

	TRISBbits.TRISB14 = 1;	// SD Card Slot in detect
	TRISBbits.TRISB15 = 1;	// SD Card Write Enable detect
	TRISBbits.TRISB7 = 0;	// SPI Slave Select --> SD Card CS
	TRISBbits.TRISB9 = 0;	// SPI clock        --> SD Card clk
	TRISBbits.TRISB8 = 0;	// SPI data output  --> SD Card in
	TRISBbits.TRISB13 = 1;	// SPI data input   <-- SD Card out


	//--- remapping I/O pin for peripheral ------------------------
	//unlock registers
	__builtin_write_OSCCONL(OSCCON & 0xBF);

	// Assign UART1	
	//RPOR2bits.RP5R = 3;		// RP5/RA0(pin 2)->UART1 TX
	//RPINR18bits.U1RXR = 6;	// RP6/RA1(pin 3)->UART1 RX

	// Assign UART2	
	RPOR2bits.RP5R = 5;		// RP5/RA0(pin 2)->UART2 TX
	RPINR19bits.U2RXR = 6;	// RP6/RA1(pin 3)->UART2 RX

	// Assign SPI1 
	RPINR20bits.SDI1R = 13; // RP13(RB13)-> SPI1 Data Input
	RPOR4bits.RP8R = 7; // RP8(RB8)-> SPI1 Data Output(7)
	RPOR4bits.RP9R = 8; // RP9(RB9)-> SPI1 Clock Output(8)

	//lock registers
	__builtin_write_OSCCONL(OSCCON | 0x40);

	//--- setting SPI for SD Card -----------------------------------
	// SD Card Slave Select(chip select) 
	LATBbits.LATB7 = 1;		// SD Card chip select to HIGH Level(deselect)


#ifdef DEBUG_MODE
    // Init UART
	UART2Init();
#endif


	//--- setting real time clock and Application Timer ------------------------------
	// see rtc.c use T1 timer
 	rtc_init();	// set the default time, 2012/07/01 09:00:00



}


/*========================================================================
	FinePix Camera Control
========================================================================*/

/*-------------------------------------------------------------------------------
   command send
-------------------------------------------------------------------------------*/

BYTE order_values[2][12] = {
	{0xc6, 0, 0, 0, 0, 0, 0,    0, 0x20, 0, 0, 0},	/* reset */
	{0xd3, 0, 0, 0, 0, 0, 0, 0x01,    0, 0, 0, 0},	/* frame request */
};

BOOL	fpix_command(BYTE deviceAddress, BYTE order)
{
	BYTE RetVal;

	while(1){
		RetVal = USBHostIssueDeviceRequest(deviceAddress,
			USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE ,
			USB_REQUEST_GET_STATUS,
			0, 0, 12, order_values[order],
			USB_DEVICE_REQUEST_SET,
			gc_DevData.clientDriverID);

		if(RetVal == USB_SUCCESS){
			USBHostTasks();
			return TRUE;
		}

		if(RetVal == USB_ENDPOINT_BUSY){
			UART2PrintString("fpix_command: USB_ENDPOINT_BUSY\r\n");
			USBHostTasks();
		}else{
			break;
		}
	}
	return FALSE;
}


/*-------------------------------------------------------------------------------
   frame read
-------------------------------------------------------------------------------*/

WORD	fpix_frame_read(BYTE deviceAddress, BYTE *buffer, DWORD length)
{
	BYTE	RetVal;

	// check busy
	while( USBHostGenericRxIsBusy(deviceAddress) == TRUE ){
		UART2PrintString("fpix_frame_read: RxIsBusy\r\n");
		USBHostTasks();
	}

	// receive request
	if((RetVal = USBHostGenericRead(deviceAddress, buffer, length)) != USB_SUCCESS ){
    	UART2PrintString( "fpix_frame_read: Device Read Error, error code=0x" );
        UART2PutHex(RetVal);
        UART2PrintString( "\r\n" );
		return 0;	// error !! received data length is 0.
	}

	// wait for transfer complete
	timer_val = 500; // 5sec
	while( USBHostGenericRxIsBusy(deviceAddress) == TRUE && camera_state == CAMERA_ATTACHED){
		if(timer_val == 0){ // 5sec
			UART2PrintString("fpix_frame_read: Transfer timed out!\r\n");
			gc_DevData.flags.rxBusy = 0;
			USBHostClearEndpointErrors(deviceAddress, FPIX_IN_EP);
			USBHostTasks();
			return 0;
		}
		USBHostTasks();
	}
	
	return USBHostGenericGetRxLength(deviceAddress);
}


// USB Host Application

//******************************************************************************
//******************************************************************************
// USB Support Functions
//******************************************************************************
//******************************************************************************

/*************************************************************************
 * Function:        USB_ApplicationEventHandler
 *
 * Preconditions:   The USB must be initialized.
 *
 * Input:           event       Identifies the bus event that occured
 *
 *                  data        Pointer to event-specific data
 *
 *                  size        Size of the event-specific data
 *
 * Output:          deviceAddress (global)
 *                  Updates device address when an attach or detach occurs.
 *
 *                  DemoState (global)
 *                  Updates the demo state as appropriate when events occur.
 *
 * Returns:         TRUE if the event was handled, FALSE if not
 *
 * Side Effects:    Event-specific actions have been taken.
 *
 * Overview:        This routine is called by the Host layer or client
 *                  driver to notify the application of events that occur.
 *                  If the event is recognized, it is handled and the
 *                  routine returns TRUE.  Otherwise, it is ignored (or
 *                  just "sniffed" and the routine returns FALSE.
 *************************************************************************/

BOOL USB_ApplicationEventHandler ( BYTE address, USB_EVENT event, void *data, DWORD size )
{
    #ifdef USB_GENERIC_SUPPORT_SERIAL_NUMBERS
        BYTE i;
    #endif

#if 0
    // Handle specific events.
    switch (event)
    {
        case EVENT_GENERIC_ATTACH:
            if (size == sizeof(GENERIC_DEVICE_ID))
            {
                deviceAddress   = ((GENERIC_DEVICE_ID *)data)->deviceAddress;
                DemoState = DEMO_STATE_GET_DEV_VERSION;
                UART2PrintString( "Generic demo device attached - event, deviceAddress=" );
                UART2PutDec( deviceAddress );
                UART2PrintString( "\r\n" );
                #ifdef USB_GENERIC_SUPPORT_SERIAL_NUMBERS
                    for (i=1; i<((GENERIC_DEVICE_ID *)data)->serialNumberLength; i++)
                    {
                        UART2PutChar( ((GENERIC_DEVICE_ID *)data)->serialNumber[i] );
                    }
                #endif
                UART2PrintString( "\r\n" );
                return TRUE;
            }
            break;

        case EVENT_GENERIC_DETACH:
            deviceAddress   = 0;
            DemoState = DEMO_INITIALIZE;
            UART2PrintString( "Generic demo device detached - event\r\n" );
            return TRUE;

        case EVENT_GENERIC_TX_DONE:  // The main state machine will poll the driver.
        case EVENT_GENERIC_RX_DONE:
            return TRUE;

        case EVENT_VBUS_REQUEST_POWER:
            // We'll let anything attach.
            return TRUE;

        case EVENT_VBUS_RELEASE_POWER:
            // We aren't keeping track of power.
            return TRUE;

        case EVENT_HUB_ATTACH:
            UART2PrintString( "\r\n***** USB Error - hubs are not supported *****\r\n" );
            return TRUE;
            break;

        case EVENT_UNSUPPORTED_DEVICE:
            UART2PrintString( "\r\n***** USB Error - device is not supported *****\r\n" );
            return TRUE;
            break;

        case EVENT_CANNOT_ENUMERATE:
            UART2PrintString( "\r\n***** USB Error - cannot enumerate device *****\r\n" );
            return TRUE;
            break;

        case EVENT_CLIENT_INIT_ERROR:
            UART2PrintString( "\r\n***** USB Error - client driver initialization error *****\r\n" );
            return TRUE;
            break;

        case EVENT_OUT_OF_MEMORY:
            UART2PrintString( "\r\n***** USB Error - out of heap memory *****\r\n" );
            return TRUE;
            break;

        case EVENT_UNSPECIFIED_ERROR:   // This should never be generated.
            UART2PrintString( "\r\n***** USB Error - unspecified *****\r\n" );
            return TRUE;
            break;

        case EVENT_SUSPEND:
        case EVENT_DETACH:
        case EVENT_RESUME:
            return TRUE;
            break;
			
        case EVENT_BUS_ERROR:
            return TRUE;
            break;

        default:
            break;
    }

#endif

    // Handle specific events.
    switch (event)
    {
        case EVENT_GENERIC_ATTACH:
            if (size == sizeof(GENERIC_DEVICE_ID))
            {
                deviceAddress   = ((GENERIC_DEVICE_ID *)data)->deviceAddress;
                camera_state = CAMERA_ATTACHED;
                UART2PrintString( "Camera attached.\r\n" );
                return TRUE;
            }

        case EVENT_GENERIC_DETACH:
            deviceAddress   = 0;
            camera_state = CAMERA_DETACHED;
            UART2PrintString( "Camera detached.\r\n" );
            return TRUE;

        case EVENT_GENERIC_TX_DONE:  // The main state machine will poll the driver.
        case EVENT_GENERIC_RX_DONE:

        case EVENT_VBUS_REQUEST_POWER:
        case EVENT_VBUS_RELEASE_POWER:
        case EVENT_HUB_ATTACH:
        case EVENT_UNSUPPORTED_DEVICE:
        case EVENT_CANNOT_ENUMERATE:
        case EVENT_CLIENT_INIT_ERROR:
        case EVENT_OUT_OF_MEMORY:
        case EVENT_UNSPECIFIED_ERROR:
        case EVENT_SUSPEND:
        case EVENT_DETACH:
        case EVENT_RESUME:
        case EVENT_BUS_ERROR:
            return TRUE;

        default:
            break;
    }

    return FALSE;

} // USB_ApplicationEventHandler




// USB Device Application

// ******************************************************************************************************
// ************** USB Callback Functions ****************************************************************
// ******************************************************************************************************
// The USB firmware stack will call the callback functions USBCBxxx() in response to certain USB related
// events.  For example, if the host PC is powering down, it will stop sending out Start of Frame (SOF)
// packets to your device.  In response to this, all USB devices are supposed to decrease their power
// consumption from the USB Vbus to <2.5mA each.  The USB module detects this condition (which according
// to the USB specifications is 3+ms of no bus activity/SOF packets) and then calls the USBCBSuspend()
// function.  You should modify these callback functions to take appropriate actions for each of these
// conditions.  For example, in the USBCBSuspend(), you may wish to add code that will decrease power
// consumption from Vbus to <2.5mA (such as by clock switching, turning off LEDs, putting the
// microcontroller to sleep, etc.).  Then, in the USBCBWakeFromSuspend() function, you may then wish to
// add code that undoes the power saving things done in the USBCBSuspend() function.

// The USBCBSendResume() function is special, in that the USB stack will not automatically call this
// function.  This function is meant to be called from the application firmware instead.  See the
// additional comments near the function.

/******************************************************************************
 * Function:        void USBCBSuspend(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Call back that is invoked when a USB suspend is detected
 *
 * Note:            None
 *****************************************************************************/
#if 0
void USBCBSuspend(void)
{
	//Example power saving code.  Insert appropriate code here for the desired
	//application behavior.  If the microcontroller will be put to sleep, a
	//process similar to that shown below may be used:
	
	//ConfigureIOPinsForLowPower();
	//SaveStateOfAllInterruptEnableBits();
	//DisableAllInterruptEnableBits();
	//EnableOnlyTheInterruptsWhichWillBeUsedToWakeTheMicro();	//should enable at least USBActivityIF as a wake source
	//Sleep();
	//RestoreStateOfAllPreviouslySavedInterruptEnableBits();	//Preferrably, this should be done in the USBCBWakeFromSuspend() function instead.
	//RestoreIOPinsToNormal();									//Preferrably, this should be done in the USBCBWakeFromSuspend() function instead.

	//IMPORTANT NOTE: Do not clear the USBActivityIF (ACTVIF) bit here.  This bit is 
	//cleared inside the usb_device.c file.  Clearing USBActivityIF here will cause 
	//things to not work as intended.	
	

    #if defined(__C30__)
    #if 0
        U1EIR = 0xFFFF;
        U1IR = 0xFFFF;
        U1OTGIR = 0xFFFF;
        IFS5bits.USB1IF = 0;
        IEC5bits.USB1IE = 1;
        U1OTGIEbits.ACTVIE = 1;
        U1OTGIRbits.ACTVIF = 1;
        Sleep();
    #endif
    #endif
}

#endif

/******************************************************************************
 * Function:        void _USB1Interrupt(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        This function is called when the USB interrupt bit is set
 *					In this example the interrupt is only used when the device
 *					goes to sleep when it receives a USB suspend command
 *
 * Note:            None
 *****************************************************************************/
#if 0
void __attribute__ ((interrupt)) _USB1Interrupt(void)
{
    #if !defined(self_powered)
        if(U1OTGIRbits.ACTVIF)
        {
            IEC5bits.USB1IE = 0;
            U1OTGIEbits.ACTVIE = 0;
            IFS5bits.USB1IF = 0;
        
            //USBClearInterruptFlag(USBActivityIFReg,USBActivityIFBitNum);
            USBClearInterruptFlag(USBIdleIFReg,USBIdleIFBitNum);
            //USBSuspendControl = 0;
        }
    #endif
}
#endif

/******************************************************************************
 * Function:        void USBCBWakeFromSuspend(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        The host may put USB peripheral devices in low power
 *					suspend mode (by "sending" 3+ms of idle).  Once in suspend
 *					mode, the host may wake the device back up by sending non-
 *					idle state signalling.
 *					
 *					This call back is invoked when a wakeup from USB suspend 
 *					is detected.
 *
 * Note:            None
 *****************************************************************************/
#if 0
void USBCBWakeFromSuspend(void)
{
	// If clock switching or other power savings measures were taken when
	// executing the USBCBSuspend() function, now would be a good time to
	// switch back to normal full power run mode conditions.  The host allows
	// a few milliseconds of wakeup time, after which the device must be 
	// fully back to normal, and capable of receiving and processing USB
	// packets.  In order to do this, the USB module must receive proper
	// clocking (IE: 48MHz clock must be available to SIE for full speed USB
	// operation).
}
#endif
/********************************************************************
 Function:        void USBCB_SOF_Handler(void)

 Description:     The USB host sends out a SOF packet to full-speed
                  devices every 1 ms. This interrupt may be useful
                  for isochronous pipes. End designers should
                  implement callback routine as necessary.

 PreCondition:    None
 
 Parameters:      None
 
 Return Values:   None
 
 Remarks:         None
 *******************************************************************/
#if 0
void USBCB_SOF_Handler(void)
{
    // No need to clear UIRbits.SOFIF to 0 here.
    // Callback caller is already doing that.
}
#endif
/*******************************************************************
 * Function:        void USBCBErrorHandler(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        The purpose of this callback is mainly for
 *                  debugging during development. Check UEIR to see
 *                  which error causes the interrupt.
 *
 * Note:            None
 *******************************************************************/
#if 0
void USBCBErrorHandler(void)
{
    // No need to clear UEIR to 0 here.
    // Callback caller is already doing that.

	// Typically, user firmware does not need to do anything special
	// if a USB error occurs.  For example, if the host sends an OUT
	// packet to your device, but the packet gets corrupted (ex:
	// because of a bad connection, or the user unplugs the
	// USB cable during the transmission) this will typically set
	// one or more USB error interrupt flags.  Nothing specific
	// needs to be done however, since the SIE will automatically
	// send a "NAK" packet to the host.  In response to this, the
	// host will normally retry to send the packet again, and no
	// data loss occurs.  The system will typically recover
	// automatically, without the need for application firmware
	// intervention.
	
	// Nevertheless, this callback function is provided, such as
	// for debugging purposes.
}

#endif
/*******************************************************************
 * Function:        void USBCBCheckOtherReq(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        When SETUP packets arrive from the host, some
 * 					firmware must process the request and respond
 *					appropriately to fulfill the request.  Some of
 *					the SETUP packets will be for standard
 *					USB "chapter 9" (as in, fulfilling chapter 9 of
 *					the official USB specifications) requests, while
 *					others may be specific to the USB device class
 *					that is being implemented.  For example, a HID
 *					class device needs to be able to respond to
 *					"GET REPORT" type of requests.  This
 *					is not a standard USB chapter 9 request, and 
 *					therefore not handled by usb_device.c.  Instead
 *					this request should be handled by class specific 
 *					firmware, such as that contained in usb_function_hid.c.
 *
 * Note:            None
 *******************************************************************/
#if 0
void USBCBCheckOtherReq(void)
{
    USBCheckMSDRequest();
}//end
#endif

/*******************************************************************
 * Function:        void USBCBStdSetDscHandler(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        The USBCBStdSetDscHandler() callback function is
 *					called when a SETUP, bRequest: SET_DESCRIPTOR request
 *					arrives.  Typically SET_DESCRIPTOR requests are
 *					not used in most applications, and it is
 *					optional to support this type of request.
 *
 * Note:            None
 *******************************************************************/
#if 0
void USBCBStdSetDscHandler(void)
{
    // Must claim session ownership if supporting this request
}//end

#endif
/*******************************************************************
 * Function:        void USBCBInitEP(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        This function is called when the device becomes
 *                  initialized, which occurs after the host sends a
 * 					SET_CONFIGURATION (wValue not = 0) request.  This 
 *					callback function should initialize the endpoints 
 *					for the device's usage according to the current 
 *					configuration.
 *
 * Note:            None
 *******************************************************************/
void USBCBInitEP(void)
{
    #if (MSD_DATA_IN_EP == MSD_DATA_OUT_EP)
        USBEnableEndpoint(MSD_DATA_IN_EP,USB_IN_ENABLED|USB_OUT_ENABLED|USB_HANDSHAKE_ENABLED|USB_DISALLOW_SETUP);
    #else
        USBEnableEndpoint(MSD_DATA_IN_EP,USB_IN_ENABLED|USB_HANDSHAKE_ENABLED|USB_DISALLOW_SETUP);
        USBEnableEndpoint(MSD_DATA_OUT_EP,USB_OUT_ENABLED|USB_HANDSHAKE_ENABLED|USB_DISALLOW_SETUP);
    #endif

    USBMSDInit();
}

/********************************************************************
 * Function:        void USBCBSendResume(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        The USB specifications allow some types of USB
 * 					peripheral devices to wake up a host PC (such
 *					as if it is in a low power suspend to RAM state).
 *					This can be a very useful feature in some
 *					USB applications, such as an Infrared remote
 *					control	receiver.  If a user presses the "power"
 *					button on a remote control, it is nice that the
 *					IR receiver can detect this signalling, and then
 *					send a USB "command" to the PC to wake up.
 *					
 *					The USBCBSendResume() "callback" function is used
 *					to send this special USB signalling which wakes 
 *					up the PC.  This function may be called by
 *					application firmware to wake up the PC.  This
 *					function will only be able to wake up the host if
 *                  all of the below are true:
 *					
 *					1.  The USB driver used on the host PC supports
 *						the remote wakeup capability.
 *					2.  The USB configuration descriptor indicates
 *						the device is remote wakeup capable in the
 *						bmAttributes field.
 *					3.  The USB host PC is currently sleeping,
 *						and has previously sent your device a SET 
 *						FEATURE setup packet which "armed" the
 *						remote wakeup capability.   
 *
 *                  If the host has not armed the device to perform remote wakeup,
 *                  then this function will return without actually performing a
 *                  remote wakeup sequence.  This is the required behavior, 
 *                  as a USB device that has not been armed to perform remote 
 *                  wakeup must not drive remote wakeup signalling onto the bus;
 *                  doing so will cause USB compliance testing failure.
 *                  
 *					This callback should send a RESUME signal that
 *                  has the period of 1-15ms.
 *
 * Note:            This function does nothing and returns quickly, if the USB
 *                  bus and host are not in a suspended condition, or are 
 *                  otherwise not in a remote wakeup ready state.  Therefore, it
 *                  is safe to optionally call this function regularly, ex: 
 *                  anytime application stimulus occurs, as the function will
 *                  have no effect, until the bus really is in a state ready
 *                  to accept remote wakeup. 
 *
 *                  When this function executes, it may perform clock switching,
 *                  depending upon the application specific code in 
 *                  USBCBWakeFromSuspend().  This is needed, since the USB
 *                  bus will no longer be suspended by the time this function
 *                  returns.  Therefore, the USB module will need to be ready
 *                  to receive traffic from the host.
 *
 *                  The modifiable section in this routine may be changed
 *                  to meet the application needs. Current implementation
 *                  temporary blocks other functions from executing for a
 *                  period of ~3-15 ms depending on the core frequency.
 *
 *                  According to USB 2.0 specification section 7.1.7.7,
 *                  "The remote wakeup device must hold the resume signaling
 *                  for at least 1 ms but for no more than 15 ms."
 *                  The idea here is to use a delay counter loop, using a
 *                  common value that would work over a wide range of core
 *                  frequencies.
 *                  That value selected is 1800. See table below:
 *                  ==========================================================
 *                  Core Freq(MHz)      MIP         RESUME Signal Period (ms)
 *                  ==========================================================
 *                      48              12          1.05
 *                       4              1           12.6
 *                  ==========================================================
 *                  * These timing could be incorrect when using code
 *                    optimization or extended instruction mode,
 *                    or when having other interrupts enabled.
 *                    Make sure to verify using the MPLAB SIM's Stopwatch
 *                    and verify the actual signal on an oscilloscope.
 *******************************************************************/
#if 0
void USBCBSendResume(void)
{
    static WORD delay_count;
    
    //First verify that the host has armed us to perform remote wakeup.
    //It does this by sending a SET_FEATURE request to enable remote wakeup,
    //usually just before the host goes to standby mode (note: it will only
    //send this SET_FEATURE request if the configuration descriptor declares
    //the device as remote wakeup capable, AND, if the feature is enabled
    //on the host (ex: on Windows based hosts, in the device manager 
    //properties page for the USB device, power management tab, the 
    //"Allow this device to bring the computer out of standby." checkbox 
    //should be checked).
    if(USBGetRemoteWakeupStatus() == TRUE) 
    {
        //Verify that the USB bus is in fact suspended, before we send
        //remote wakeup signalling.
        if(USBIsBusSuspended() == TRUE)
        {
            USBMaskInterrupts();
            
            //Clock switch to settings consistent with normal USB operation.
            USBCBWakeFromSuspend();
            USBSuspendControl = 0; 
            USBBusIsSuspended = FALSE;  //So we don't execute this code again, 
                                        //until a new suspend condition is detected.

            //Section 7.1.7.7 of the USB 2.0 specifications indicates a USB
            //device must continuously see 5ms+ of idle on the bus, before it sends
            //remote wakeup signalling.  One way to be certain that this parameter
            //gets met, is to add a 2ms+ blocking delay here (2ms plus at 
            //least 3ms from bus idle to USBIsBusSuspended() == TRUE, yeilds
            //5ms+ total delay since start of idle).
            delay_count = 3600U;        
            do
            {
                delay_count--;
            }while(delay_count);
            
            //Now drive the resume K-state signalling onto the USB bus.
            USBResumeControl = 1;       // Start RESUME signaling
            delay_count = 1800U;        // Set RESUME line for 1-13 ms
            do
            {
                delay_count--;
            }while(delay_count);
            USBResumeControl = 0;       //Finished driving resume signalling

            USBUnmaskInterrupts();
        }
    }
}
#endif

/*******************************************************************
 * Function:        BOOL USER_USB_CALLBACK_EVENT_HANDLER(
 *                        USB_EVENT event, void *pdata, WORD size)
 *
 * PreCondition:    None
 *
 * Input:           USB_EVENT event - the type of event
 *                  void *pdata - pointer to the event data
 *                  WORD size - size of the event data
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        This function is called from the USB stack to
 *                  notify a user application that a USB event
 *                  occured.  This callback is in interrupt context
 *                  when the USB_INTERRUPT option is selected.
 *
 * Note:            None
 *******************************************************************/
BOOL USER_USB_CALLBACK_EVENT_HANDLER(USB_EVENT event, void *pdata, WORD size)
{

#if 0
    switch((INT)event)
    {
        case EVENT_TRANSFER:
            //Add application specific callback task or callback function here if desired.
            break;
        case EVENT_SOF:
            //USBCB_SOF_Handler();
            break;
        case EVENT_SUSPEND:
            //USBCBSuspend();
            break;
        case EVENT_RESUME:
            //USBCBWakeFromSuspend();
            break;
        case EVENT_CONFIGURED: 
            USBCBInitEP();
            break;
        case EVENT_SET_DESCRIPTOR:
            //USBCBStdSetDscHandler();
            break;
        case EVENT_EP0_REQUEST:
            //USBCBCheckOtherReq();
			USBCheckMSDRequest();
            break;
        case EVENT_BUS_ERROR:
            //USBCBErrorHandler();
            break;
        case EVENT_TRANSFER_TERMINATED:
            //Add application specific callback task or callback function here if desired.
            //The EVENT_TRANSFER_TERMINATED event occurs when the host performs a CLEAR
            //FEATURE (endpoint halt) request on an application endpoint which was 
            //previously armed (UOWN was = 1).  Here would be a good place to:
            //1.  Determine which endpoint the transaction that just got terminated was 
            //      on, by checking the handle value in the *pdata.
            //2.  Re-arm the endpoint if desired (typically would be the case for OUT 
            //      endpoints).

            //Check if the host recently did a clear endpoint halt on the MSD OUT endpoint.
            //In this case, we want to re-arm the MSD OUT endpoint, so we are prepared
            //to receive the next CBW that the host might want to send.
            //Note: If however the STALL was due to a CBW not valid condition, 
            //then we are required to have a persistent STALL, where it cannot 
            //be cleared (until MSD reset recovery takes place).  See MSD BOT 
            //specs v1.0, section 6.6.1.
            if(MSDWasLastCBWValid() == FALSE)
            {
                //Need to re-stall the endpoints, for persistent STALL behavior.
    			USBStallEndpoint(MSD_DATA_IN_EP, IN_TO_HOST);
      			USBStallEndpoint(MSD_DATA_OUT_EP, OUT_FROM_HOST);                 
            }
            else
            {   
                //Check if the host cleared halt on the bulk out endpoint.  In this
                //case, we should re-arm the endpoint, so we can receive the next CBW.
                if((USB_HANDLE)pdata == USBGetNextHandle(MSD_DATA_OUT_EP, OUT_FROM_HOST))
                {
                    USBMSDOutHandle = USBRxOnePacket(MSD_DATA_OUT_EP, (BYTE*)&msd_cbw, MSD_OUT_EP_SIZE);
                }    
            }    
            break;
        default:
            break;
    }

#endif

    switch((INT)event)
    {
        case EVENT_CONFIGURED: 
            USBCBInitEP();
            break;

        case EVENT_EP0_REQUEST:
			USBCheckMSDRequest();
            break;

        case EVENT_TRANSFER_TERMINATED:
            if(MSDWasLastCBWValid() == FALSE)
            {
                //Need to re-stall the endpoints, for persistent STALL behavior.
    			USBStallEndpoint(MSD_DATA_IN_EP, IN_TO_HOST);
      			USBStallEndpoint(MSD_DATA_OUT_EP, OUT_FROM_HOST);                 
            }
            else
            {   
                //Check if the host cleared halt on the bulk out endpoint.  In this
                //case, we should re-arm the endpoint, so we can receive the next CBW.
                if((USB_HANDLE)pdata == USBGetNextHandle(MSD_DATA_OUT_EP, OUT_FROM_HOST))
                {
                    USBMSDOutHandle = USBRxOnePacket(MSD_DATA_OUT_EP, (BYTE*)&msd_cbw, MSD_OUT_EP_SIZE);
                }    
            }    
            break;

        default:
            break;
    }

    return TRUE; 
}




/*************************************************************************
 * EOF main.c
 */



